import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:osam/osam.dart';

class OsamNavigator<P extends PresenterWithNavigation<BaseState>> extends StatefulWidget {
  final Map<String, WidgetBuilder> screens;
  final bool rootNavigator;
  final bool isMaterial;

  const OsamNavigator({Key key, this.screens, this.rootNavigator = false, this.isMaterial = false})
      : super(key: key);

  @override
  _OsamNavigatorState createState() => _OsamNavigatorState<P>();
}

class _OsamNavigatorState<P extends PresenterWithNavigation<BaseState>> extends State<OsamNavigator>
    with AfterLayoutMixin {
  final List<String> _routes = [];

  Map<String, WidgetBuilder> get screens => widget.screens;

  @override
  Widget build(BuildContext context) =>
      screens[context.presenter<P>().routeStream.initialData.first](context);

  void _silentPop(PresenterWithNavigation presenter) {
    if (presenter.routeStream.initialData.length > 1) {
      _routes.removeLast();
      presenter.pop();
    }
  }

  PageRoute pageRoute({WidgetBuilder builder, RouteSettings settings}) => widget.isMaterial
      ? MaterialPageRoute(builder: builder, settings: settings)
      : CupertinoPageRoute(builder: builder, settings: settings);

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context, rootNavigator: widget.rootNavigator);
    final presenter = context.presenter<P>();
    final initialRoutes = presenter.routeStream.initialData;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final firstRoute = initialRoutes.first;
      navigator.pushReplacement(pageRoute(
          builder: ((ctx) => screens[firstRoute])(context),
          settings: const RouteSettings(isInitialRoute: true)));
      final others = initialRoutes.sublist(1, initialRoutes.length);
      others.forEach((route) {
        navigator.push(pageRoute(builder: ((ctx) => screens[route])(context))
          ..popped.then((_) => _silentPop(presenter)));
      });
    }
    presenter.routeStream.stream.listen((routes) {
      final routesDifferentNumber = routes.length - _routes.length;
      if (routesDifferentNumber == 0 && routes.last != _routes.last) {
        navigator.pushReplacement(pageRoute(builder: ((ctx) {
          return screens[routes.last](context);
        }))
          ..popped.then((_) => _silentPop(presenter)));
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = routes.sublist(_routes.length, routes.length);
        pushRoutes.forEach((route) {
          navigator.push(pageRoute(builder: ((ctx) => screens[route])(context))
            ..popped.then((_) => _silentPop(presenter)));
        });
      } else if (routesDifferentNumber < 0) {
        _routes.first != routes.first
            ? navigator.pushAndRemoveUntil(pageRoute(builder: ((ctx) {
                return screens[routes.first](context);
              })), (route) => false)
            : _routes.sublist(routes.length, _routes.length).forEach((route) async {
                navigator.pop();
              });
      }
      _routes.clear();
      _routes.addAll(routes);
    });
  }
}

class OsamTabNavigator<P extends PresenterWithTabNavigation<BaseState>> extends StatelessWidget {
  final Map<String, Widget> navigators;
  final Map<String, Widget> tabs;
  final BoxDecoration tabDecoration;
  final _navigatorKeys = <String, GlobalKey<NavigatorState>>{};

  OsamTabNavigator({Key key, this.navigators, this.tabDecoration, this.tabs}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final presenter = context.presenter<P>();
    if (_navigatorKeys.isEmpty) {
      _navigatorKeys.addAll(navigators.map((k, v) => MapEntry(k, GlobalKey<NavigatorState>())));
    }
    return OsamBuilder<String>(
      key: const ValueKey('tab_ob'),
      stream: presenter.tabStream,
      builder: (ctx, snapshot) {
        return Stack(
          children: <Widget>[
            IndexedStack(
              index: navigators.keys.toList().indexOf(snapshot.data),
              children: navigators.keys.map((tab) {
                return _OsamNavigatorPage(
                  key: ValueKey(tab.toString() + '_onp'),
                  navKey: _navigatorKeys[tab],
                  child: navigators[tab],
                );
              }).toList(),
            ),
            Positioned(
              bottom: 0,
              child: SafeArea(
                bottom: true,
                child: Container(
                  decoration: tabDecoration,
                  padding: const EdgeInsets.symmetric(vertical: 5),
                  constraints: BoxConstraints(
                    maxWidth: MediaQuery.of(context).size.width,
                    minHeight: 48,
                  ),
                  child: Row(
                    children: navigators.keys
                        .map((tab) => Expanded(
                              child: GestureDetector(
                                child: Container(
                                  child: tabs[tab],
                                  alignment: Alignment.center,
                                  color: Colors.transparent,
                                ),
                                onTap: () {
                                  if (snapshot.data != tab) {
                                    presenter.changeTab(tab);
                                  } else {
//                                    final currentPresenter = navigators[snapshot.data].presenter;
//                                    final routes = currentPresenter.routeStream.initialData;
//                                    if (routes.length > 1) {
//                                      currentPresenter.pushAndRemoveUntil(routes.first);
//                                    }
                                  }
                                },
                              ),
                            ))
                        .toList(),
                  ),
                ),
              ),
            )
          ],
        );
      },
    );
  }
}

class _OsamNavigatorPage extends StatelessWidget {
  final GlobalKey<NavigatorState> navKey;
  final Widget child;

  const _OsamNavigatorPage({Key key, this.navKey, this.child}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Navigator(
      key: navKey,
      onGenerateRoute: (RouteSettings settings) => MaterialPageRoute(
        settings: const RouteSettings(isInitialRoute: true),
        builder: (BuildContext context) => child,
      ),
    );
  }
}
