import 'package:flutter/material.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/domain/store/store.dart';
import 'package:osam/osam.dart';
import 'package:osam/presentation/presenter.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';

class StoreProvider extends InheritedWidget {
  StoreProvider({Key key, @required this.child, @required this.store, this.persist})
      : super(
            key: key,
            child: (persist?.isPersistEnabled ?? false) ? _LifecycleWrapper(child: child) : child);
  final Store store;
  final PersistRepository persist;

  @override
  final Widget child;

  static StoreProvider of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType(aspect: StoreProvider);

  @override
  bool updateShouldNotify(StoreProvider oldWidget) => false;
}

class _LifecycleWrapper extends StatefulWidget {
  final Widget child;

  const _LifecycleWrapper({Key key, this.child}) : super(key: key);

  @override
  _LifecycleWrapperState createState() => _LifecycleWrapperState();
}

class _LifecycleWrapperState extends State<_LifecycleWrapper> with WidgetsBindingObserver {
  @override
  void initState() {
    WidgetsBinding.instance.addObserver(this);
    super.initState();
  }

  @override
  Widget build(BuildContext context) => widget.child;

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.paused) {
      final storeProvider = StoreProvider.of(context);
      storeProvider.persist.save(storeProvider.store.state);
    }
  }
}

mixin _PresenterProviderSingleChildWidget on SingleChildWidget {}

class PresenterProvider<P extends Presenter<BaseState>> extends SingleChildStatelessWidget
    with _PresenterProviderSingleChildWidget {
  final P presenter;
  final Widget child;

  const PresenterProvider({Key key, @required this.presenter, this.child})
      : super(key: key, child: child);

  static P of<P extends Presenter<BaseState>>(BuildContext context) =>
      Provider.of<P>(context, listen: false);

  @override
  Widget buildWithChild(BuildContext context, Widget child) {
    final store = StoreProvider.of(context).store;
    assert(store != null);
    if (store == null)
      debugPrint('insure you provided store by StoreProvider in the start of your widget tree');
    return Provider(
      key: ValueKey('PresenterProvider: %${key.toString()}'),
      create: (_) => presenter
        // ignore: invalid_use_of_protected_member
        ..store = store
        ..init(),
      child: child,
      dispose: (ctx, presenter) => presenter.dispose(),
    );
  }
}

class MultiPresenterProvider extends StatelessWidget {
  final List<SingleChildWidget> providers;
  final Widget child;

  const MultiPresenterProvider({Key key, @required this.providers, @required this.child})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MultiProvider(
      providers: providers,
      key: ValueKey('PresenterMultiProvider: %${key.toString()}'),
      child: child,
    );
  }
}

extension PresenterProviderExtension on BuildContext {
  P presenter<P extends Presenter<BaseState>>() => PresenterProvider.of<P>(this);
}
